import re
import os
import typing

from enum import Enum
from asyncio.tasks import Task
from urllib.parse import unquote
from collections import namedtuple
from weakref import WeakKeyDictionary
from flask.sessions import SecureCookieSession

from toolkit import global_cache_classproperty, load
from toolkit.settings import SettingsLoader, Settings, FrozenSettings

from .exceptions import Readonly

Cookie = typing.NewType('Cookie', str)
# 用于标记已知名字的表单字段
FormParam = typing.NewType("FormParam", str)
Session = typing.NewType("Session", SecureCookieSession)
# 这种类型用于标记有且只有文件的表单
MultiPartForm = typing.NewType("MultiPartForm", dict)
# 这种类型用于标记不存在文件的表单
UrlEncodeForm = typing.NewType("UrlEncodeForm", dict)
DummyFlaskApp = namedtuple(
    "DummyFlaskApp",
    "session_cookie_name,secret_key,permanent_session_lifetime,config")

# 全局settings对象，使用init_settings来初始化
settings = FrozenSettings(Settings())


class File(object):
    mime_type_regex = re.compile(rb"Content-Type: (\S*)")
    content_length_regex = re.compile(rb"Content-Length: (\d*)")
    disposition_regex = re.compile(
        rb'Content-Disposition: form-data(?:; name\*?=\"?'
        rb'(?:(?P<name_enc>[\w\-]+?)'
        rb'\'(?P<name_lang>\w*)\')?'
        rb'(?P<name>[^\";]*)\"?)?.*?(?:; filename\*?=\"?'
        rb'(?:(?P<enc>[\w\-]+?)'
        rb'\'(?P<lang>\w*)\')?'
        rb'(?P<filename>[^\"]*?)\"?)?(?:$|\r\n)')

    def __init__(self, stream, receive, boundary, name,
                 filename, mimetype, content_length):
        self.content_length = content_length
        self.mimetype = mimetype
        self.receive = receive
        self.filename = filename
        self.name = name
        self.stream = stream
        self.tmpboundary = b"\r\n--" + boundary
        self.boundary_len = len(self.tmpboundary)
        self._last = b""
        self._size = 0
        self.body_iter = self._iter_content()

    def __aiter__(self):
        return self.body_iter

    def __str__(self):
        return f"<{self.__class__.__name__} " \
               f"name={self.name} " \
               f"filename={self.filename} >"

    __repr__ = __str__

    def iter_content(self):
        return self.body_iter

    async def _iter_content(self):
        stream = self.stream
        while True:
            # 如果存在read过程中剩下的，则直接返回
            if self._last:
                read = self._last
                self._last = b""
                yield read
                continue

            index = self.stream.body.find(self.tmpboundary)
            if index != -1:
                # 找到分隔线，返回分隔线前的数据
                # 并将分隔及分隔线后的数据返回给stream
                read, stream.body = stream.body[:index], stream.body[index:]
                self._size += len(read)
                yield read
                if self._last:
                    read = self._last
                    self._last = b""
                    yield read
                break
            else:
                assert not self.stream.closed, "Content not complete!"
                # 若没有找到分隔线，为了防止分隔线被读取了一半
                # 选择只返回少于分隔线长度的部分body
                read = stream.body[:-self.boundary_len]
                stream.body = stream.body[-self.boundary_len:]
                self._size += len(read)
                yield read
                await self.get_message(self.receive, stream)

    async def read(self, size=10240):
        """
        推荐直接迭代File对象，性能较好

        :param size:
        :return:
        """
        read = b""
        assert size > 0, (999, "Read size must > 0")
        while len(read) < size:
            try:
                buffer = await self.body_iter.asend(None)
            except StopAsyncIteration:
                return read
            read = read + buffer
            read, self._last = read[:size], read[size:]

        return read

    @staticmethod
    async def get_message(receive, stream):
        message = await receive()

        assert message['type'] == 'http.request', \
            f"Unexpected ASGI message type: {message['type']}."

        if not message.get('more_body', False):
            stream.closed = True
        stream.body += message.get("body", b"")

    def tell(self):
        return self._size - len(self._last)

    @classmethod
    async def from_boundary(cls, stream, receive, boundary):
        tmp_boundary = b"--" + boundary
        while not stream.closed:
            await cls.get_message(receive, stream)

            if b"\r\n\r\n" in stream.body and tmp_boundary in stream.body or \
                    stream.closed:
                break

        return cls(
            stream, receive, boundary, *cls.get_headers(stream, tmp_boundary))

    @classmethod
    def get_headers(cls, stream, tmp_boundary):
        end_boundary = tmp_boundary + b"--"
        body = stream.body
        index = body.find(tmp_boundary)
        if index == body.find(end_boundary):
            raise StopAsyncIteration
        body = body[index + len(tmp_boundary):]
        split_index = body.find(b"\r\n\r\n")
        header_str, body = body[:split_index], body[split_index + 4:]
        headers = cls.parse(header_str)
        stream.body = body
        return headers

    @classmethod
    def parse(cls, header_str):
        groups = cls.disposition_regex.search(header_str).groupdict()

        filename = groups["filename"] and unquote(groups["filename"].decode())
        if groups["enc"]:
            filename = filename.encode().decode(groups["enc"].decode())

        name = groups["name"] and unquote(groups["name"].decode())
        if groups["name_enc"]:
            name = name.encode().decode(groups["name_enc"].decode())

        mth = cls.mime_type_regex.search(header_str)
        mimetype = mth and mth.group(1).decode()
        mth = cls.content_length_regex.search(header_str) or 0
        content_length = int(mth and mth.group(1))
        assert name, "FileStream iterated without File consumed. "
        return name, filename, mimetype, content_length


class FileStream(object):

    def __init__(self, receive, boundary):
        self.receive = receive
        self.boundary = boundary
        self.body = b""
        self.closed = False

    def __aiter__(self):
        return self

    async def __anext__(self):
        return await File.from_boundary(self, self.receive, self.boundary)


class InheritType(Enum):
    DUPLICATE = 0  # 重名且类型相符，在参数列表和赋值中不体现，在super中体现
    OVERWRITE = 1  # 重名但是类型不同，全部位置要体现，但是需要父类的参数改名字
    NORMAL = 2  # 正常的情况


class Inject(object):

    def __init__(self, type, default):
        self.annotation = self.type = type
        self.name = type.__name__.lower()
        self.default = default

    def __set__(self, instance, value):
        raise Readonly(f"Readonly object of {self.type.__name__}.")

    def __get__(self, instance, cls):
        if instance:
            return instance.__dict__[self.name]
        else:
            return self


class InjectManager(object):

    def __init__(self, default=None):
        self.default = default

    def __lshift__(self, other):
        return Inject(other, self.default)

    def __call__(self, *, default=None):
        return InjectManager(default)


inject = InjectManager()


class SettingsMixin(object):
    settings_path = None

    @global_cache_classproperty
    def settings(cls):
        """
        这个应该在app加载时调用，这样project_path才准确。
        由于有些项目可能在模块作用域就需要使用settings中的配置，
        所以调用应该在项目模块加载之前。

        :return:
        """
        return SettingsLoader().load(
            SettingsMixin.settings_path or "settings",
            default={"PROJECT_PATH": os.path.abspath(os.getcwd())})


def init_settings(settings_path):
    """
    初始化settings
    :param settings_path:
    :return:
    """
    SettingsMixin.settings_path = settings_path
    settings._json.update(SettingsMixin.settings._json)


class Local(object):
    """
    实现类似threadlocal的效果
    """
    coroutinelocal = WeakKeyDictionary()

    _default = dict(scope="apistar.server.asgi.ASGIScope")

    @global_cache_classproperty
    def local_variable(cls):
        """
        存储配置的可注入的上下文变量及其类型
        :return:
        """
        lv = cls._default.copy()
        lv.update(settings.get("LOCAL_VARIABLE", dict()))
        var_types = dict()
        for k, v in lv.items():
            var_types[k] = load(v)
        return var_types

    @staticmethod
    def current_task():
        return Task.current_task()

    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError as e:
            raise AttributeError(e)

    def __getitem__(self, item):
        task = self.current_task()
        if task is None:
            raise KeyError("current task not found!")
        return self.coroutinelocal[self.current_task()][item]

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            return default

    def __setattr__(self, key, value):
        raise Readonly("Readonly object: Local!")

    @classmethod
    def set_default(cls, name, val_path):
        cls._default[name] = val_path


coroutinelocal = Local()



import json
import asyncio

from abc import ABCMeta
from collections.abc import Mapping, MutableMapping

from apistar.exceptions import ConfigurationError, ValidationError

from . import validators
from ..persistence import PersistentMeta
from ..helper import TypeEncoder, add_success_callback


class TypeMetaclass(ABCMeta):

    def __new__(mcs, name, bases, attrs):
        properties = []
        for key, value in list(attrs.items()):
            if key in ['keys', 'items', 'values', 'get', "clear",
                       'validator', "setdefault", "pop", "popitem"]:
                msg = (
                    'Cannot use reserved name "%s" on Type "%s", as it '
                    'clashes with the class interface.'
                )
                raise ConfigurationError(msg % (key, name))

            elif hasattr(value, 'validate'):
                attrs.pop(key)
                properties.append((key, value))

        # If this class is subclassing another Type, add that Type's properties.
        # Note that we loop over the bases in reverse. This is necessary in order
        # to maintain the correct order of properties.
        for base in reversed(bases):
            if hasattr(base, 'validator'):
                properties = [
                    (key, base.validator.properties[key]) for key
                    in base.validator.properties
                    if key not in attrs
                ] + properties

        properties = sorted(
            properties,
            key=lambda item: item[1]._creation_counter
        )

        required = [
            key for key, value in properties
            if not value.has_default()
        ]
        attrs['_creation_counter'] = validators.Validator._creation_counter
        validators.Validator._creation_counter += 1
        cls = super(TypeMetaclass, mcs).__new__(mcs, name, bases, attrs)

        cls.validator = validators.Object(
            def_name=name,
            properties=properties,
            required=required,
            additional_properties=None,
            model=cls
        )

        return cls


class Type(MutableMapping, metaclass=TypeMetaclass):

    def __init__(self, *args, **kwargs):
        definitions = None
        allow_coerce = False
        force_format = False

        if args:
            assert len(args) == 1
            definitions = kwargs.pop('definitions', definitions)
            allow_coerce = kwargs.pop('allow_coerce', allow_coerce)
            force_format = kwargs.pop('force_format', force_format)
            assert not kwargs

            if args[0] is None or isinstance(args[0], (str, bytes, bool, int, float, list)):
                raise ValidationError('Must be an object.')
            elif isinstance(args[0], Mapping):
                # Instantiated with a dict.
                value = args[0]
            else:
                # Instantiated with an object instance.
                value = {}
                for key, val in self.validator.properties.items():
                    v = getattr(args[0], key, None)
                    if v:
                        value[key] = v
        else:
            # Instantiated with keyword arguments.
            value = kwargs
        object.__setattr__(self, 'allow_coerce', allow_coerce)
        object.__setattr__(self, '_dict', dict(value))
        object.__setattr__(self, 'formatted', False)

        if force_format:
            self.format(self.allow_coerce)

    def format(self, allow_coerce=False):
        object.__setattr__(self, 'allow_coerce', allow_coerce)
        if not self.formatted:
            object.__setattr__(self, '_dict', self.validator.validate(
                self._dict, allow_coerce=self.allow_coerce))
            object.__setattr__(self, 'formatted', True)

    @classmethod
    def validate(cls, value, definitions=None, allow_coerce=False, force_format=True):
        return cls(value,
                   definitions=definitions,
                   allow_coerce=allow_coerce,
                   force_format=force_format)

    def __repr__(self):
        if self.formatted:
            pair = self.items()
        else:
            pair = self._dict.items()
        args = ['%s=%s' % (key, repr(value)) for key, value in pair]
        arg_string = ', '.join(args)
        return '<%s(%s)>' % (self.__class__.__name__, arg_string)

    def __setattr__(self, key, value):
        if key not in self.validator.properties:
            raise AttributeError('Invalid attribute "%s"' % key)
        value = self.validator.properties[key].validate(value)
        self._dict[key] = value

    def __setitem__(self, key, value):
        if key not in self.validator.properties:
            raise KeyError('Invalid key "%s"' % key)
        value = self.validator.properties[key].validate(value)
        self._dict[key] = value

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        del self._dict[key]
        if self.formatted:
            self._dict[key] = self.validator.properties[key].validate(None)

    def __getattr__(self, key):
        try:
            return self._dict[key]
        except (KeyError,):
            raise AttributeError('Invalid attribute "%s"' % key)

    def __getitem__(self, key):

        validator = self.validator.properties.get(key)
        try:
            value = self._dict[key]
        except KeyError as e:
            try:
                if validator:
                    value = validator.get_default()
                    self[key] = value
                else:
                    raise e
            except AssertionError:
                raise e

        if hasattr(validator, 'formatter') and validator.formatter:
            return validator.formatter.to_string(value)

        return value

    def __contains__(self, item):
        return item in self._dict

    def __len__(self):
        return len(self._dict)

    def __iter__(self):
        return iter(self._dict)

    @classmethod
    def init(cls, **kwargs):
        for k, v in kwargs.items():
            setattr(cls, k, v)

    def update(self, **kwargs):
        for k, v in kwargs.items():
            if k in self.validator.properties:
                setattr(self, k, v)
            else:
                self._dict[k] = v

    def reformat(self, field_name, value=None, allow_coerce=False):
        if value is None:
            value = self._dict.get(field_name)
        val = self.__class__.validator.properties[field_name].validate(
                    value, allow_coerce=allow_coerce)
        setattr(self, field_name, val)
        return val

    def to_dict(self):
        return json.loads(json.dumps(self, cls=TypeEncoder))

    to_json = to_dict


class PersistentTypeMeta(PersistentMeta, TypeMetaclass):
    pass


class PersistentType(Type, metaclass=PersistentTypeMeta):
    """
    拥有持久化能力的Type
    """
    pass


class AsyncType(Type):
    """
    获取属性时为空时，自动load数据
    """
    def __getattr__(self, item):
        """
        实现__getattr__是为了点获取属性时变成异步操作。
        :param item:
        :return:
        """
        loop = asyncio.get_event_loop()
        val_future = loop.create_future()
        properties = super().__getattribute__("validator").properties.keys()

        if item in properties:
            # 对于item存在于properties中的属性，则通过父类去获取
            # 若能获取到，则将future置为done。否则进行异步加载
            try:
                val = super().__getattr__(item)
                val_future.set_result(val)
            except AttributeError:
                self._load(val_future, item)
        else:
            # 不item不存在于properties中，则证明找不到item且item
            # 不是字段。直接将future设置异常。
            val_future.set_exception(AttributeError(item))
        return val_future

    def _load(self, val_future, item):
        def set_val(doc_or_exc):
            try:
                self._dict.update(doc_or_exc)
                val_future.set_result(doc_or_exc[item])
            except TypeError:
                val_future.set_exception(doc_or_exc)
            except Exception as e:
                val_future.set_exception(e)

        task = add_success_callback(self.load(), set_val)
        asyncio.get_event_loop().create_task(task)

    async def load(self):
        return NotImplemented



import os
import time
import typing
import mimetypes

from apistar.http import Response, StrMapping, StrPairs

from apistellar.helper import parse_date


class FileResponse(Response):
    def __init__(self,
                 content: typing.Any,
                 filename: str = None,
                 status_code: int = 200,
                 headers: typing.Union[StrMapping, StrPairs] = None,
                 exc_info=None,
                 media_type: str = "auto",
                 charset: str = 'utf-8',
                 download: bool = True,
                 req_headers: typing.Union[StrMapping, StrPairs] = None,
    ):
        self.filename = filename
        self.media_type = media_type or self.media_type
        self.charset = charset
        self.download = download
        self.req_headers = req_headers
        super(FileResponse, self).__init__(content, status_code, headers, exc_info)

    def render(self, content: typing.Any):
        if hasattr(content, "read"):
            if not self.filename:
                self.filename = getattr(content, "filename", None) \
                                or os.path.basename(getattr(content, "name", ""))

            return content

        if isinstance(content, bytes):
            return content
        elif isinstance(content, str) and self.charset is not None:
            return content.encode(self.charset)

        valid_types = "bytes" if self.charset is None else "string or bytes"
        raise AssertionError(
            "%s content must be %s. Got %s." %
            (self.__class__.__name__, valid_types, type(content).__name__)
        )

    def set_default_headers(self):
        if 'Content-Length' not in self.headers:
            if hasattr(self.content, "read"):
                assert False, "File like object need specify Content-Length."
            self.headers['Content-Length'] = str(len(self.content))

        assert self.filename, "filename must specify."

        if self.media_type == 'auto':
            self.media_type, encoding = mimetypes.guess_type(self.filename)
            if encoding:
                self.headers['Content-Encoding'] = encoding

        if self.media_type:
            if self.media_type.startswith('text/') and \
                    self.charset and 'charset' not in self.media_type:
                self.media_type += f'; charset={self.charset}'
            self.headers['Content-Type'] = self.media_type

        if self.download:
            self.headers['Content-Disposition'] = f'attachment; ' \
                                                  f'filename="{self.filename}"'
        if "Last-Modified" not in self.headers:
            self.headers['Last-Modified'] = \
                time.strftime("%a, %d %b %Y %H:%M:%S GMT")

        if self.req_headers and "If-Modified-Since" in self.req_headers:
            if_modified_since = parse_date(
                self.req_headers["If-Modified-Since"].split(";")[0].strip())
        else:
            if_modified_since = None

        mtime = parse_date(self.headers['Last-Modified'])
        if if_modified_since is not None and mtime is not None and \
                if_modified_since >= int(mtime):
            self.status_code = 304



import logging

from toolkit import cache_property
from abc import ABC, abstractmethod


class Solo(ABC):
    """
    独立任务程序子任务接口定义
    """
    name = None

    @cache_property
    def logger(self):
        logging.basicConfig(level=logging.DEBUG)
        return logging.getLogger(self.name or self.__class__.__name__)

    def __init__(self, **kwargs):
        self.tasks = []

    @abstractmethod
    async def setup(self, *args, **kwargs):
        """
        任务初始化
        :param args:
        :param kwargs:
        :return:
        """

    @abstractmethod
    async def teardown(self, *args, **kwargs):
        """
        资源回收
        :param args:
        :param kwargs:
        :return:
        """

    @abstractmethod
    async def run(self, *args, **kwargs):
        """
        任务执行
        :param args:
        :param kwargs:
        :return:
        """

    @classmethod
    def enrich_parser(cls, sub_parser):
        """
        自定义命令行参数
        :param sub_parser:
        :return:
        """